/**
 * Copyright (C) 2013  Dionis Argiri
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package hqlcriteria;

import com.google.common.collect.*;
import hqlcriteria.expression.Operator;
import hqlcriteria.expression.Restriction;
import hqlcriteria.expression.Restrictions;
import hqlcriteria.expression.BindParam;
import hqlcriteria.expression.Expression;
import hqlcriteria.transformers.ResultTransformer;
import org.hibernate.Query;
import org.hibernate.Session;

import java.util.*;
import java.util.Map.Entry;

/**
 * @author dionis
 */
public class HQLCriteria {
    private final String rootEntityName;
    private final String rootEntityAlias;
    private ResultTransformer resultTransformer;
    private ObjectGraphLoadingMode objectGraphLoadingMode = ObjectGraphLoadingMode.CONSTRAINED;
    /**
     * Association of path to alias
     */
    private final BiMap<String, String> aliases = HashBiMap.create();
    /**
     * Association path(probably including alias) to join.
     */
    private final Map<String, JoinType> joins = Maps.newHashMap();
    /**
     * Alias or association path to fetch profile.
     */
    private final Map<String, FetchProfile> fetchProfiles = Maps.newLinkedHashMap();
    /**
     * Restrictions.
     */
    private final Set<Expression> restrictions = Sets.newLinkedHashSet();
    /**
     * Which entities to select
     */
    private final Map<String, String> fromEntities = Maps.newLinkedHashMap();
    /**
     * Path to ordering
     */
    private final Map<String, Order> ordering = Maps.newLinkedHashMap();

    public static HQLCriteria create(String rootEntityName,
                                     String rootEntityAlias) {
        return new HQLCriteria(rootEntityName, rootEntityAlias);
    }

    public static HQLCriteria create(Class<?> rootEntityClazz,
                                     String rootEntityAlias) {
        if (rootEntityClazz == null) throw new IllegalArgumentException();
        return new HQLCriteria(rootEntityClazz.getSimpleName(), rootEntityAlias);
    }

    private HQLCriteria(String rootEntityName, String rootEntityAlias) {
        if (rootEntityName == null) throw new IllegalArgumentException();
        if (rootEntityAlias == null) throw new IllegalArgumentException();
        this.rootEntityName = rootEntityName;
        this.rootEntityAlias = rootEntityAlias;
    }

    public HQLCriteria setResultTransformer(ResultTransformer resultTransformer) {
        this.resultTransformer = resultTransformer;
        return this;
    }

    public HQLCriteria createAlias(String association, String alias, JoinType joinType) {
        if (association == null) throw new IllegalArgumentException();
        if (alias == null) throw new IllegalArgumentException();
        if (joinType == null) throw new IllegalArgumentException();
        aliases.put(association, alias);
        joins.put(alias, joinType);
        if (!fetchProfiles.containsKey(alias)) {
            fetchProfiles.put(alias, FetchProfile.LAZY);// lazy by default
        }
        return this;
    }

    public HQLCriteria createAlias(String association, String alias) {
        return createAlias(association, alias, JoinType.INNER);
    }

    public HQLCriteria setFetchProfile(String association,
                                       FetchProfile fetchProfile) {
        return setFetchProfile(association, fetchProfile, JoinType.INNER);
    }

    public HQLCriteria setFetchProfile(String association, FetchProfile fetchProfile, JoinType joinType) {
        if (association == null) throw new IllegalArgumentException();
        if (fetchProfile == null) throw new IllegalArgumentException();
        if (joinType == null) throw new IllegalArgumentException();
        fetchProfiles.put(association, fetchProfile);
        if (!joins.containsKey(association)) {
            joins.put(association, joinType);
        }
        return this;
    }

    public HQLCriteria add(Restriction restriction) {
        if (restriction == null) throw new IllegalArgumentException();
        restrictions.add(restriction);
        return this;
    }

    public HQLCriteria add(Operator restriction) {
        if (restriction == null) throw new IllegalArgumentException();
        restrictions.add(restriction);
        return this;
    }

    @SuppressWarnings("rawtypes")
    public List list(Session session) {
        HQLCriteriaBuildContext hqlCriteriaBuildContext = new HQLCriteriaBuildContext();
        Query query = session.createQuery(buildHQLQueryString(hqlCriteriaBuildContext));
        for (BindParam bindParam : hqlCriteriaBuildContext.getBindParams()) {
            bindParam.setParameterToQuery(query);
        }

        List result = query.list();
        if (resultTransformer != null) {
            return resultTransformer.transform(result);
        } else {
            return result;
        }
    }

    public String buildHQLQueryString() {
        return buildHQLQueryString(new HQLCriteriaBuildContext());
    }

    private String buildHQLQueryString(HQLCriteriaBuildContext hqlCriteriaBuildContext) {
        switch (objectGraphLoadingMode) {
            case CONSTRAINED:
                final StringBuilder queryBuilder = new StringBuilder();
                appendRootEntityAliasSelect(queryBuilder);
                appendJoins(queryBuilder);
                appendRestrictionsIfPossible(queryBuilder, hqlCriteriaBuildContext);
                appendOrdering(queryBuilder);
                return queryBuilder.toString();
            case FULL:
                return transform().buildHQLQueryString(hqlCriteriaBuildContext);
            default:
                throw new UnsupportedOperationException("Can't use object graph loading mode " + objectGraphLoadingMode.name());

        }
    }

    private void appendOrdering(StringBuilder queryBuilder) {
        if(!ordering.isEmpty()){
            queryBuilder.append(" order by ");
            Iterator<Entry<String, Order>> iterator = ordering.entrySet().iterator();
            while (iterator.hasNext()){
                Entry<String, Order> orderEntry = iterator.next();
                queryBuilder.append(orderEntry.getKey()).append(" ").append(toHQLOrderingString(orderEntry.getValue()));
                if(iterator.hasNext()){
                    queryBuilder.append(", ");
                }
            }
        }
    }

    private String toHQLOrderingString(Order order) {
        switch (order){
            case ASCENDING:
                return "asc";
            case DESCENDING:
                return "desc";
            default:
                throw new UnsupportedOperationException("Unsupportet ordering type "+order.name());
        }
    }

    public HQLCriteria transform() {
        BiMap<String, String> aliases = HashBiMap.create(this.aliases);
        Map<String, JoinType> joins = Maps.newHashMap(this.joins);
        Map<String, FetchProfile> fetchProfileMap = Maps.newHashMap(this.fetchProfiles);
        HashMap<String, Order> ordering = Maps.newHashMap(this.ordering);
        String oldRootAlias = this.rootEntityAlias;
        BiMap<String, String> aliasToPath = aliases.inverse();
        Map<String, String> oldAliasToNewAlias = Maps.newLinkedHashMap();
        String newRootAlias = createNewAlias(oldRootAlias, aliasToPath);

        oldAliasToNewAlias.put(oldRootAlias, newRootAlias);
        for (String s : aliasToPath.keySet()) {
            oldAliasToNewAlias.put(s, createNewAlias(s, aliasToPath));
        }

        for (Map.Entry<String, String> oldToNew : oldAliasToNewAlias.entrySet()) {
            replaceAliases(oldToNew.getKey(), oldToNew.getValue(), aliases, joins, fetchProfileMap, ordering);
        }

        HQLCriteria transformed = HQLCriteria.create(this.rootEntityName, newRootAlias);
        transformed.aliases.putAll(aliases);
        transformed.aliases.putAll(this.aliases);
        transformed.joins.putAll(this.joins);
        transformed.joins.putAll(joins);
        transformed.ordering.putAll(this.ordering);
        transformed.ordering.putAll(ordering);
        for (String association : this.fetchProfiles.keySet()) {
            transformed.fetchProfiles.put(association, FetchProfile.LAZY);
        }
        transformed.fetchProfiles.putAll(fetchProfileMap);
        transformed.restrictions.addAll(this.restrictions);
        transformed.restrictions.add(Restrictions.eq(rootEntityAlias, newRootAlias));
        transformed.fromEntities.put(rootEntityName, rootEntityAlias);

        return transformed;
    }

    private void replaceAliases(String oldRootAlias, String newRootAlias, BiMap<String, String> aliases, Map<String, JoinType> joins, Map<String, FetchProfile> fetchProfileMap, Map<String, Order> ordering) {
        BiMap<String, String> aliasMap = replaceAliasesInAliases(oldRootAlias, newRootAlias, aliases);
        aliases.clear();
        aliases.putAll(aliasMap);
        Map<String, JoinType> joinTypeMap = replaceAliasesInJoins(oldRootAlias, newRootAlias, joins);
        joins.clear();
        joins.putAll(joinTypeMap);
        Map<String, FetchProfile> fetchProfileMap1 = replaceAliasesInFetchProfiles(oldRootAlias, newRootAlias, fetchProfileMap);
        fetchProfileMap.clear();
        fetchProfileMap.putAll(fetchProfileMap1);
        Map<String, Order> orderingMapResult = replaceAliasesInOrdering(oldRootAlias, newRootAlias, ordering);
        ordering.clear();
        ordering.putAll(orderingMapResult);
    }

    private Map<String, Order> replaceAliasesInOrdering(String oldAlias, String replacement, Map<String, Order> ordering) {
        Map<String, Order> result = Maps.newHashMapWithExpectedSize(ordering.size());
        //replace root alias
        for (Entry<String, Order> entry : ordering.entrySet()) {
            String aliasKey = entry.getKey();
            if (aliasKey.startsWith(oldAlias + ".")) {
                result.put(replacement + aliasKey.substring(aliasKey.indexOf(".")), entry.getValue());
            } else if (entry.getKey().equals(oldAlias)) {
                result.put(replacement, entry.getValue());
            } else {
                result.put(aliasKey, entry.getValue());
            }
        }
        return result;
    }

    private Map<String, FetchProfile> replaceAliasesInFetchProfiles(String oldAlias, String replacement, Map<String, FetchProfile> fetchProfileMap) {
        Map<String, FetchProfile> result = Maps.newHashMapWithExpectedSize(fetchProfileMap.size());
        //replace root alias
        for (Entry<String, FetchProfile> entry : fetchProfileMap.entrySet()) {
            String aliasKey = entry.getKey();
            if (aliasKey.startsWith(oldAlias + ".")) {
                result.put(replacement + aliasKey.substring(aliasKey.indexOf(".")), entry.getValue());
            } else if (entry.getKey().equals(oldAlias)) {
                result.put(replacement, entry.getValue());
            } else {
                result.put(aliasKey, entry.getValue());
            }
        }
        return result;
    }

    private Map<String, JoinType> replaceAliasesInJoins(String oldAlias, String replacement, Map<String, JoinType> joins) {
        Map<String, JoinType> result = Maps.newHashMapWithExpectedSize(joins.size());
        //replace root alias
        for (Map.Entry<String, JoinType> entry : joins.entrySet()) {
            String aliasKey = entry.getKey();
            if (aliasKey.startsWith(oldAlias + ".")) {
                result.put(replacement + aliasKey.substring(aliasKey.indexOf(".")), entry.getValue());
            } else if (entry.getKey().equals(oldAlias)) {
                result.put(replacement, entry.getValue());
            } else {
                result.put(aliasKey, entry.getValue());
            }
        }
        return result;
    }

    private BiMap<String, String> replaceAliasesInAliases(String oldAlias, String replacement, BiMap<String, String> aliases) {

        BiMap<String, String> result = HashBiMap.create(aliases.size());
        //replace root alias
        for (Map.Entry<String, String> entry : aliases.entrySet()) {
            String aliasKey = entry.getKey();
            if (aliasKey.startsWith(oldAlias + ".")) {
                result.put(replacement + aliasKey.substring(aliasKey.indexOf(".")), entry.getValue());
            } else if (entry.getValue().equals(oldAlias)) {
                result.put(aliasKey, replacement);
            } else {
                result.put(aliasKey, entry.getValue());
            }
        }
        return result;
    }

    private String createNewAlias(String alias, Map<String, String> aliasToPath) {
        String newAlias;
        do {
            newAlias = alias + "_";
        } while (aliasToPath.containsKey(newAlias));
        return newAlias;
    }

    public HQLCriteria copy() {
        HQLCriteria hqlCriteria = HQLCriteria.create(rootEntityName, rootEntityAlias);
        hqlCriteria.aliases.putAll(this.aliases);
        hqlCriteria.fetchProfiles.putAll(this.fetchProfiles);
        hqlCriteria.joins.putAll(this.joins);
        hqlCriteria.restrictions.addAll(this.restrictions);
        hqlCriteria.fromEntities.putAll(this.fromEntities);
        hqlCriteria.ordering.putAll(this.ordering);
        hqlCriteria.setObjectGraphLoadingMode(this.objectGraphLoadingMode);
        hqlCriteria.setResultTransformer(this.resultTransformer);
        return hqlCriteria;
    }

    private void appendRestrictionsIfPossible(final StringBuilder queryBuilder, HQLCriteriaBuildContext hqlCriteriaBuildContext) {
        if (!restrictions.isEmpty()) {
            queryBuilder.append(" where ");
            appendRestrictions(queryBuilder, hqlCriteriaBuildContext);
        }
    }

    private void appendJoins(final StringBuilder queryBuilder) {
        for (final Entry<String, FetchProfile> fetchProfile : fetchProfiles.entrySet()) {
            final String associationOrAliasToAssociation = fetchProfile.getKey();

            final FetchProfile profile = fetchProfile.getValue();
            final boolean isAliasToAssociation = aliases.containsValue(associationOrAliasToAssociation);
            final JoinType joinType = joins.get(associationOrAliasToAssociation);

            queryBuilder.append(" ");
            queryBuilder.append(toHQLJoinString(joinType));
            String hqlFetchProfileString = toHQLFetchProfileString(profile);
            if (hqlFetchProfileString.length()>0) {
                queryBuilder.append(" ");
                queryBuilder.append(hqlFetchProfileString);
            }

            final String associationPath;
            final String alias;
            if (isAliasToAssociation) {
                associationPath = aliases.inverse().get(
                        associationOrAliasToAssociation);
                alias = associationOrAliasToAssociation;
            } else {
                associationPath = associationOrAliasToAssociation;
                alias = "";
            }

            queryBuilder.append(" ");
            queryBuilder.append(associationPath);
            if (alias.length()>0) {
                queryBuilder.append(" ");
                queryBuilder.append(alias);
            }
        }
    }

    private void appendRootEntityAliasSelect(final StringBuilder queryBuilder) {
        queryBuilder
                .append("select ").append(rootEntityAlias)
                .append(" from ").append(rootEntityName).append(" ").append(rootEntityAlias);
        for (Entry<String, String> from : fromEntities.entrySet()) {
            String entityName = from.getKey();
            String entityAlias = from.getValue();
            queryBuilder.append(", ").append(entityName).append(" ").append(entityAlias).append(" ");
        }
    }

    private void appendRestrictions(StringBuilder queryBuilder, HQLCriteriaBuildContext hqlCriteriaBuildContext) {
        Expression prevRestriction = null;
        for (Expression restriction : restrictions) {
            if (prevRestriction != null) {
                queryBuilder.append(" and ");
            }
            queryBuilder.append(restriction.apply(hqlCriteriaBuildContext));
            prevRestriction = restriction;
        }
    }

    private String toHQLFetchProfileString(FetchProfile fetchMode) {
        switch (fetchMode) {
            case EAGER:
                return "fetch";
            case LAZY:
                return "";
            default:
                throw new UnsupportedOperationException(fetchMode.name()
                        + " can not be converted to hql fetch strategy string");
        }
    }

    private String toHQLJoinString(JoinType joinType) {
        switch (joinType) {
            case INNER:
                return "inner join";
            case LEFT:
                return "left join";
            case RIGHT:
                return "right join";
            default:
                throw new UnsupportedOperationException(joinType.name()
                        + " can not be converted to HQL join string");
        }
    }

    public HQLCriteria setObjectGraphLoadingMode(ObjectGraphLoadingMode objectGraphLoadingMode) {
        if (objectGraphLoadingMode == null) throw new IllegalArgumentException();
        this.objectGraphLoadingMode = objectGraphLoadingMode;
        return this;
    }

    public HQLCriteria addOrdering(String path, Order order) {
        if(order == null) throw new IllegalArgumentException();
        ordering.put(path, order);
        return this;
    }
}
