/**
 * Copyright (C) 2013  Dionis Argiri
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package hqlcriteria.expression;

import hqlcriteria.HQLCriteriaBuildContext;
import org.hibernate.Query;

import java.io.Serializable;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Calendar;
import java.util.Date;

/**
 * @author dionis
 */
public abstract class BindParam<T> implements Expression {
    private final T value;
    private final String paramName;

    public BindParam(String paramName, T value) {
        this.value = value;
        this.paramName = paramName;
    }

    public String getParamName() {
        return paramName;
    }

    public T getValue() {
        return value;
    }
    
    public abstract void setParameterToQuery(Query query);

    public String apply(HQLCriteriaBuildContext hqlCriteriaBuildContext) {
        hqlCriteriaBuildContext.addBindParam(this);
        return ":" + paramName;
    }

    public static class ObjectBind extends BindParam<Object> {

        public ObjectBind(String paramName, Object value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setParameter(getParamName(),getValue());
        }

    }
    
    public static class BigDecimalBind extends BindParam<BigDecimal> {


        public BigDecimalBind(String paramName, BigDecimal value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setBigDecimal(getParamName(),getValue());
        }

    }

    public static class BigIntegerBind extends BindParam<BigInteger> {


        public BigIntegerBind(String paramName, BigInteger value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setBigInteger(getParamName(), getValue());
        }
    }

    public static class BinaryBind extends BindParam<byte[]> {


        public BinaryBind(String paramName, byte[] value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setBinary(getParamName(),getValue());
        }
    }

    public static class BooleanBind extends BindParam<Boolean> {


        public BooleanBind(String paramName, Boolean value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setBoolean(getParamName(),getValue());
        }
    }

    public static class ByteBind extends BindParam<Byte> {


        public ByteBind(String paramName, Byte value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setByte(getParamName(),getValue());
        }
    }

    public static class CalendarBind extends BindParam<Calendar> {


        public CalendarBind(String paramName, Calendar value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setCalendar(getParamName(),getValue());
        }
    }

    public static class CalendarDateBind extends BindParam<Calendar> {


        public CalendarDateBind(String paramName, Calendar value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setCalendarDate(getParamName(),getValue());
        }
    }

    public static class CharacterBind extends BindParam<Character> {


        public CharacterBind(String paramName, Character value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setCharacter(getParamName(), getValue());
        }
    }

    public static class DateBind extends BindParam<Date> {


        public DateBind(String paramName, Date value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setDate(getParamName(),getValue());
        }
    }

    public static class DoubleBind extends BindParam<Double> {


        public DoubleBind(String paramName, Double value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setDouble(getParamName(),getValue());
        }
    }

    public static class FloatBind extends BindParam<Float> {


        public FloatBind(String paramName, Float value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setFloat(getParamName(),getValue());
        }
    }

    public static class IntegerBind extends BindParam<Integer> {

        public IntegerBind(String paramName, Integer value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setInteger(getParamName(),getValue());
        }
    }

    public static class LongBind extends BindParam<Long> {


        public LongBind(String paramName, Long value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setLong(getParamName(),getValue());
        }
    }

    public static class ShortBind extends BindParam<Short> {


        public ShortBind(String paramName, Short value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setShort(getParamName(),getValue());
        }
    }

    public static class SerializableBind extends BindParam<Serializable> {


        public SerializableBind(String paramName, Serializable value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setSerializable(getParamName(),getValue());
        }
    }

    public static class StringBind extends BindParam<String> {


        public StringBind(String paramName, String value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setString(getParamName(),getValue());
        }
    }

    public static class TextBind extends BindParam<String> {

        public TextBind(String paramName, String value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setText(getParamName(),getValue());
        }
    }

    public static class TimeBind extends BindParam<Date> {


        public TimeBind(String paramName, Date value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setTime(getParamName(),getValue());
        }
    }

    public static class TimestampBind extends BindParam<Date> {


        public TimestampBind(String paramName, Date value) {
            super(paramName, value);
        }

        @Override
        public void setParameterToQuery(Query query) {
            query.setTimestamp(getParamName(),getValue());
        }
    }


}
