/**
 * Copyright (C) 2013  Dionis Argiri
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package hqlcriteria.expression;

import hqlcriteria.HQLCriteriaBuildContext;

/**
 * @author dionis
 */
public class Restrictions {
	private static class HQLRestriction implements Restriction {
		private final String hql;
		
		public HQLRestriction(String hql) {
			this.hql = hql;
		}

		public String apply(HQLCriteriaBuildContext hqlCriteriaBuildContext) {
			return hql;
		}
		
	}
	
	private static class Equal implements Restriction {
		private final Expression left;
		private final Expression right;

		public Equal(Expression left, Expression right) {
			this.left = left;
			this.right = right;
		}

		public String apply(HQLCriteriaBuildContext hqlCriteriaBuildContext) {
			return left.apply(hqlCriteriaBuildContext) + "=" + right.apply(hqlCriteriaBuildContext);
		}
	}

	private static class LessThen implements Restriction {
		private final Expression left;
		private final Expression right;

		public LessThen(Expression left, Expression right) {
			this.left = left;
			this.right = right;
		}

		public String apply(HQLCriteriaBuildContext hqlCriteriaBuildContext) {
			return left.apply(hqlCriteriaBuildContext) + "<" + right.apply(hqlCriteriaBuildContext);
		}
	}

	private static class GreaterThen implements Restriction {
		private final Expression left;
		private final Expression right;

		public GreaterThen(Expression left, Expression right) {
			this.left = left;
			this.right = right;
		}

		public String apply(HQLCriteriaBuildContext hqlCriteriaBuildContext) {
			return left.apply(hqlCriteriaBuildContext) + ">" + right.apply(hqlCriteriaBuildContext);
		}
	}

	private static class LessThenOrEqual implements Restriction {
		private final Expression left;
		private final Expression right;

		public LessThenOrEqual(Expression left, Expression right) {
			this.left = left;
			this.right = right;
		}

		public String apply(HQLCriteriaBuildContext hqlCriteriaBuildContext) {
			return left.apply(hqlCriteriaBuildContext) + "<" + right.apply(hqlCriteriaBuildContext);
		}
	}

	private static class GreaterThenOrEqual implements Restriction {
		private final Expression left;
		private final Expression right;

		public GreaterThenOrEqual(Expression left, Expression right) {
			this.left = left;
			this.right = right;
		}

		public String apply(HQLCriteriaBuildContext hqlCriteriaBuildContext) {
			return left.apply(hqlCriteriaBuildContext) + ">" + right.apply(hqlCriteriaBuildContext);
		}
	}

    private static class LessThenGreaterThen implements Restriction {
        private final Expression left;
        private final Expression right;

        public LessThenGreaterThen(Expression left, Expression right) {
            this.left = left;
            this.right = right;
        }

        public String apply(HQLCriteriaBuildContext hqlCriteriaBuildContext) {
            return left.apply(hqlCriteriaBuildContext) + "<>" + right.apply(hqlCriteriaBuildContext);
        }
    }

    private static class NotEqual implements Restriction {
        private final Expression left;
        private final Expression right;

        public NotEqual(Expression left, Expression right) {
            this.left = left;
            this.right = right;
        }

        public String apply(HQLCriteriaBuildContext hqlCriteriaBuildContext) {
            return left.apply(hqlCriteriaBuildContext) + "!=" + right.apply(hqlCriteriaBuildContext);
        }
    }

    private static class Like implements Restriction {
        private final Expression left;
        private final Expression right;

        public Like(Expression left, Expression right) {
            this.left = left;
            this.right = right;
        }

        public String apply(HQLCriteriaBuildContext hqlCriteriaBuildContext) {
            return left.apply(hqlCriteriaBuildContext) + "!=" + right.apply(hqlCriteriaBuildContext);
        }
    }
	
	public static Restriction hqlRestriction(String hql){
		return new HQLRestriction(hql);
	}

	public static Restriction eq(Expression left, Expression right) {
		return new Equal(left, right);
	}

	public static Restriction eq(String left, Expression right) {
		return new Equal(Value.of(left), right);
	}

	public static Restriction eq(Expression left, String right) {
		return new Equal(left, Value.of(right));
	}

	public static Restriction eq(String left, String right) {
		return new Equal(Value.of(left), Value.of(right));
	}

	public static Restriction lt(Expression left, Expression right) {
		return new LessThen(left, right);
	}

	public static Restriction lt(String left, Expression right) {
		return new LessThen(Value.of(left), right);
	}

	public static Restriction lt(Expression left, String right) {
		return new LessThen(left, Value.of(right));
	}

	public static Restriction lt(String left, String right) {
		return new LessThen(Value.of(left), Value.of(right));
	}

	public static Restriction gt(Expression left, Expression right) {
		return new GreaterThen(left, right);
	}

	public static Restriction gt(String left, Expression right) {
		return new GreaterThen(Value.of(left), right);
	}

	public static Restriction gt(Expression left, String right) {
		return new GreaterThen(left, Value.of(right));
	}

	public static Restriction gt(String left, String right) {
		return new GreaterThen(Value.of(left), Value.of(right));
	}

	public static Restriction ltOrEq(Expression left, Expression right) {
		return new LessThenOrEqual(left, right);
	}

	public static Restriction ltOrEq(String left, Expression right) {
		return new LessThenOrEqual(Value.of(left), right);
	}

	public static Restriction ltOrEq(Expression left, String right) {
		return new LessThenOrEqual(left, Value.of(right));
	}

	public static Restriction ltOrEq(String left, String right) {
		return new LessThenOrEqual(Value.of(left), Value.of(right));
	}

	public static Restriction gtOrEq(Expression left, Expression right) {
		return new GreaterThenOrEqual(left, right);
	}

	public static Restriction gtOrEq(String left, Expression right) {
		return new GreaterThenOrEqual(Value.of(left), right);
	}

	public static Restriction gtOrEq(Expression left, String right) {
		return new GreaterThenOrEqual(left, Value.of(right));
	}

	public static Restriction gtOrEq(String left, String right) {
		return new GreaterThenOrEqual(Value.of(left), Value.of(right));
	}

    public static Restriction ltGt(Expression left, Expression right) {
        return new LessThenGreaterThen(left, right);
    }

    public static Restriction ltGt(String left, Expression right) {
        return new LessThenGreaterThen(Value.of(left), right);
    }

    public static Restriction ltGt(Expression left, String right) {
        return new LessThenGreaterThen(left, Value.of(right));
    }

    public static Restriction ltGt(String left, String right) {
        return new LessThenGreaterThen(Value.of(left), Value.of(right));
    }

    public static Restriction ne(Expression left, Expression right) {
        return new NotEqual(left, right);
    }

    public static Restriction ne(String left, Expression right) {
        return new NotEqual(Value.of(left), right);
    }

    public static Restriction ne(Expression left, String right) {
        return new NotEqual(left, Value.of(right));
    }

    public static Restriction ne(String left, String right) {
        return new NotEqual(Value.of(left), Value.of(right));
    }

    public static Restriction like(String left, String right) {
        return new Like(Value.of(left), Value.of(right));
    }
}
